/*
   (c) Copyright 2001-2011  The world wide DirectFB Open Source Community (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de>,
              Sven Neumann <neo@directfb.org>,
              Ville Syrjälä <syrjala@sci.fi> and
              Claudio Ciccani <klan@users.sf.net>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include "SaWMan.h"

#include <directfb_util.h>

#include <direct/debug.h>
#include <direct/mem.h>
#include <direct/memcpy.h>
#include <direct/messages.h>

#include <fusion/conf.h>

#include <core/core.h>

D_DEBUG_DOMAIN( DirectFB_SaWMan, "DirectFB/SaWMan", "DirectFB SaWMan" );

/*********************************************************************************************************************/

DFBResult
SaWMan_RegisterProcess(
                    SaWMan                                    *obj,
                    SaWManProcessFlags                         flags,
                    s32                                        pid,
                    u32                                        fusion_id,
                    SaWManProcess                            **ret_process
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        return ISaWManWM_Real__RegisterProcess( obj, flags, pid, fusion_id, ret_process );
    }

    return ISaWManWM_Requestor__RegisterProcess( obj, flags, pid, fusion_id, ret_process );
}

DFBResult
SaWMan_Start(
                    SaWMan                                    *obj,
                    const u8                                  *name,
                    u32                                        name_len,
                    s32                                       *ret_pid
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        return ISaWManWM_Real__Start( obj, name, name_len, ret_pid );
    }

    return ISaWManWM_Requestor__Start( obj, name, name_len, ret_pid );
}

DFBResult
SaWMan_Stop(
                    SaWMan                                    *obj,
                    s32                                        pid
)
{
    if (!fusion_config->secure_fusion || dfb_core_is_master( core_dfb )) {
        return ISaWManWM_Real__Stop( obj, pid );
    }

    return ISaWManWM_Requestor__Stop( obj, pid );
}

/*********************************************************************************************************************/

static FusionCallHandlerResult
SaWMan_Dispatch( int           caller,   /* fusion id of the caller */
                     int           call_arg, /* optional call parameter */
                     void         *ptr, /* optional call parameter */
                     unsigned int  length,
                     void         *ctx,      /* optional handler context */
                     unsigned int  serial,
                     void         *ret_ptr,
                     unsigned int  ret_size,
                     unsigned int *ret_length )
{
    SaWMan *obj = (SaWMan*) ctx;
    SaWManDispatch__Dispatch( obj, caller, call_arg, ptr, length, ret_ptr, ret_size, ret_length );

    return FCHR_RETURN;
}

void SaWMan_Init_Dispatch(
                    CoreDFB              *core,
                    SaWMan               *obj,
                    FusionCall           *call
)
{
    fusion_call_init3( call, SaWMan_Dispatch, obj, core->world );
}

void  SaWMan_Deinit_Dispatch(
                    FusionCall           *call
)
{
     fusion_call_destroy( call );
}

/*********************************************************************************************************************/


DFBResult
ISaWManWM_Requestor__RegisterProcess( SaWMan *obj,
                    SaWManProcessFlags                         flags,
                    s32                                        pid,
                    u32                                        fusion_id,
                    SaWManProcess                            **ret_process
)
{
    DFBResult           ret;
    SaWManProcess *process = NULL;
    SaWManRegisterProcess       *block = (SaWManRegisterProcess*) alloca( sizeof(SaWManRegisterProcess) );
    SaWManRegisterProcessReturn  return_block;

    D_DEBUG_AT( DirectFB_SaWMan, "ISaWManWM_Requestor::%s()\n", __FUNCTION__ );

    D_ASSERT( ret_process != NULL );

    block->flags = flags;
    block->pid = pid;
    block->fusion_id = fusion_id;

    ret = (DFBResult) SaWMan_Call( obj, FCEF_NONE, _SaWMan_RegisterProcess, block, sizeof(SaWManRegisterProcess), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: SaWMan_Call( SaWMan_RegisterProcess ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: SaWMan_RegisterProcess failed!\n", __FUNCTION__ );
         return return_block.result;
    }


    ret = (DFBResult) SaWManProcess_Catch( core_dfb, return_block.process_id, &process );
    if (ret) {
         D_DERROR( ret, "%s: Catching process by ID %u failed!\n", __FUNCTION__, return_block.process_id );
         return ret;
    }

    *ret_process = process;

    return DFB_OK;
}


DFBResult
ISaWManWM_Requestor__Start( SaWMan *obj,
                    const u8                                  *name,
                    u32                                        name_len,
                    s32                                       *ret_pid
)
{
    DFBResult           ret;
    SaWManStart       *block = (SaWManStart*) alloca( sizeof(SaWManStart) + sizeof(u8) * name_len );
    SaWManStartReturn  return_block;

    D_DEBUG_AT( DirectFB_SaWMan, "ISaWManWM_Requestor::%s()\n", __FUNCTION__ );


    block->name_len = name_len;
    direct_memcpy( (char*) (block + 1), name, sizeof(u8) * name_len );

    ret = (DFBResult) SaWMan_Call( obj, FCEF_NONE, _SaWMan_Start, block, sizeof(SaWManStart) + sizeof(u8) * name_len, &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: SaWMan_Call( SaWMan_Start ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: SaWMan_Start failed!\n", __FUNCTION__ );
         return return_block.result;
    }

    *ret_pid = return_block.pid;

    return DFB_OK;
}


DFBResult
ISaWManWM_Requestor__Stop( SaWMan *obj,
                    s32                                        pid
)
{
    DFBResult           ret;
    SaWManStop       *block = (SaWManStop*) alloca( sizeof(SaWManStop) );
    SaWManStopReturn  return_block;

    D_DEBUG_AT( DirectFB_SaWMan, "ISaWManWM_Requestor::%s()\n", __FUNCTION__ );


    block->pid = pid;

    ret = (DFBResult) SaWMan_Call( obj, FCEF_NONE, _SaWMan_Stop, block, sizeof(SaWManStop), &return_block, sizeof(return_block), NULL );
    if (ret) {
        D_DERROR( ret, "%s: SaWMan_Call( SaWMan_Stop ) failed!\n", __FUNCTION__ );
        return ret;
    }

    if (return_block.result) {
         D_DERROR( return_block.result, "%s: SaWMan_Stop failed!\n", __FUNCTION__ );
         return return_block.result;
    }


    return DFB_OK;
}

/*********************************************************************************************************************/

static DFBResult
__SaWManDispatch__Dispatch( SaWMan *obj,
                                FusionID      caller,
                                int           method,
                                void         *ptr,
                                unsigned int  length,
                                void         *ret_ptr,
                                unsigned int  ret_size,
                                unsigned int *ret_length )
{
    D_UNUSED
    DFBResult ret;


    switch (method) {
        case _SaWMan_RegisterProcess: {
    SaWManProcess *process = NULL;
            D_UNUSED
            SaWManRegisterProcess       *args        = (SaWManRegisterProcess *) ptr;
            SaWManRegisterProcessReturn *return_args = (SaWManRegisterProcessReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_SaWMan, "=-> SaWMan_RegisterProcess\n" );

            return_args->result = ISaWManWM_Real__RegisterProcess( obj, args->flags, args->pid, args->fusion_id, &process );
            if (return_args->result == DFB_OK) {
                SaWManProcess_Throw( process, caller, &return_args->process_id );
            }

            *ret_length = sizeof(SaWManRegisterProcessReturn);

            return DFB_OK;
        }

        case _SaWMan_Start: {
            D_UNUSED
            SaWManStart       *args        = (SaWManStart *) ptr;
            SaWManStartReturn *return_args = (SaWManStartReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_SaWMan, "=-> SaWMan_Start\n" );

            return_args->result = ISaWManWM_Real__Start( obj, (u8*) ((char*)(args + 1)), args->name_len, &return_args->pid );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(SaWManStartReturn);

            return DFB_OK;
        }

        case _SaWMan_Stop: {
            D_UNUSED
            SaWManStop       *args        = (SaWManStop *) ptr;
            SaWManStopReturn *return_args = (SaWManStopReturn *) ret_ptr;

            D_DEBUG_AT( DirectFB_SaWMan, "=-> SaWMan_Stop\n" );

            return_args->result = ISaWManWM_Real__Stop( obj, args->pid );
            if (return_args->result == DFB_OK) {
            }

            *ret_length = sizeof(SaWManStopReturn);

            return DFB_OK;
        }

    }

    return DFB_NOSUCHMETHOD;
}
/*********************************************************************************************************************/

DFBResult
SaWManDispatch__Dispatch( SaWMan *obj,
                                FusionID      caller,
                                int           method,
                                void         *ptr,
                                unsigned int  length,
                                void         *ret_ptr,
                                unsigned int  ret_size,
                                unsigned int *ret_length )
{
    DFBResult ret;

    D_DEBUG_AT( DirectFB_SaWMan, "SaWManDispatch::%s()\n", __FUNCTION__ );

    ret = __SaWManDispatch__Dispatch( obj, caller, method, ptr, length, ret_ptr, ret_size, ret_length );

    return ret;
}
